/*
	BromScript - On Runtime Scripting Language.
    Copyright (C) 2012 - 2014  Alex Brouwer (Bromvlieg: http://brom.4o3.nl/)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "../Objects/Function.h"

using namespace Scratch;

namespace BromScript {
	Function::Function(Instance* bs, CString* strtbl):CurrentSourceFileLine(-1),Code(null),ForceReturn(false),IsCpp(false),Parent(null),BromScript(bs),CppFunc(null),CurrentThisObject(null),Locals(new Table(bs)),StringTable(strtbl){
	}

	Function::~Function(){
		if (this->Code != null)
			delete[] this->Code;

		delete this->Locals;
	}

	void Function::SetVar(CString key, Variable* var, bool global){
		if (global){
			BromScript::Function* func = this;
			while(true){
				if (func == null){
					this->BromScript->Globals->Set(key, var);
					return;
				}else if (func->Locals->HasKey(key)){
					func->Locals->Set(key, var);
					return;
				}

				func = func->Parent;
			}
		}

		this->Locals->Set(key, var);
	}


	Variable* Function::GetVar(CString key){
		if (key == "this" && this->CurrentThisObject != null)
			return this->CurrentThisObject;

		BromScript::Function* func = this;
		while(true){
			if (func == null){
				return this->BromScript->Globals->Get(key);
			}else if (func->Locals->HasKey(key)){
				return func->Locals->Get(key);
			}

			func = func->Parent;
		}
	}

	Variable* Function::Run(){
		ArgumentData* args = new ArgumentData();
		args->BromScript = this->BromScript;
		args->Caller = this;
		args->Count = 0;

		Variable* ret = this->Run(args);
		delete args;

		return ret;
	}
	
	Variable* Function::Run(Variable* thisobject){
		Variable* oldthis = this->CurrentThisObject;
		this->CurrentThisObject = thisobject;

		ArgumentData* args = new ArgumentData();
		args->BromScript = this->BromScript;
		args->Caller = this;
		
		Variable* ret = this->Run(args);
		delete args;
		
		this->CurrentThisObject = oldthis;
		return ret;
	}

	Variable* Function::Run(List<Variable*>* vars){
		ArgumentData* args = new ArgumentData();
		args->BromScript = this->BromScript;
		args->Caller = this;
		args->SetVariableData(vars);

		Variable* ret = this->Run(args);
		delete args;

		return ret;
	}

	Variable* Function::Run(List<Variable*>* vars, Variable* thisobject){
		Variable* oldthis = this->CurrentThisObject;
		this->CurrentThisObject = thisobject;

		Variable* ret = this->Run(vars);

		this->CurrentThisObject = oldthis;
		return ret;
	}

	Variable* Function::Run(ArgumentData* args){
		if (this->BromScript->GetCurrentStackSize() >= 257){
			args->Error(CString::Format("Stack error, cannot call '%s'", this->Name.str_szBuffer));
			return this->BromScript->GC.RegisterVariable();
		}

		this->ForceReturn = false;
		
		if (this->IsCpp){
			args->ThisObject = this->CurrentThisObject;

			this->BromScript->EnterStack(this);
			Variable* ret = this->CppFunc(this->BromScript, args);
			this->BromScript->LeaveStack();

			if (ret == null) ret = this->BromScript->GC.RegisterVariable();
			else this->BromScript->GC.RegisterVariable(ret);
			
			return ret;
		}

		for (int i = 0; i < args->Count && i < this->Parameters.Count; i++)
			this->SetVar(this->Parameters[i], args->GetVariable(i), false);

		ExecuteData* data = new ExecuteData();
		data->BromScript = this->BromScript;
		data->Reader = new ByteReader();
		data->Reader->Func = this;
		data->Function = this;

		int oldline = this->CurrentSourceFileLine;
		this->CurrentSourceFileLine = -1;

		this->BromScript->EnterStack(this);
		Variable* ret = this->InternalRun(data);
		this->BromScript->LeaveStack();

		this->CurrentSourceFileLine = oldline;

		delete data;
		return ret;
	}
	
	Variable* Function::Run(ArgumentData* args, Variable* thisobject){
		Variable* oldthis = this->CurrentThisObject;
		this->CurrentThisObject = thisobject;

		Variable* ret = this->Run(args);

		this->CurrentThisObject = oldthis;
		return ret;
	}

	Variable* Function::InternalRun(ExecuteData* data){
		while(true){
			if (data->HasReturnValue || data->Breaking || data->Continueing || data->Function->ForceReturn)
				return data->Returning;

			if (data->BromScript->KillScriptThreaded)
				return data->BromScript->GC.RegisterVariable();

			Misc::ExecFuncs b = (Misc::ExecFuncs)data->Reader->ReadByte();
			
			switch(b){
			case Misc::ExecFuncs::GetVar: return Executer::GetVar(data); break;
			case Misc::ExecFuncs::GetTblIndex:
				if (data->FuncCallback) return Executer::GetTableIndex(data);
				else Executer::GetTableIndex(data);
				break; 
			
			case Misc::ExecFuncs::SetTblIndex:Executer::SetTableIndex(data); break; 

			case Misc::ExecFuncs::SetVar: Executer::SetVar(data); break;
			case Misc::ExecFuncs::LSetVar: Executer::LSetVar(data); break;
			
			case Misc::ExecFuncs::PostIncrement: if (data->FuncCallback) return Executer::PostIncrement(data); else Executer::PostIncrement(data); break; 
			case Misc::ExecFuncs::PreIncrement: if (data->FuncCallback) return Executer::PreIncrement(data); else Executer::PreIncrement(data); break; 
			case Misc::ExecFuncs::PostDecrement: if (data->FuncCallback) return Executer::PostDecrement(data); else Executer::PostDecrement(data); break; 
			case Misc::ExecFuncs::PreDecrement: if (data->FuncCallback) return Executer::PreDecrement(data); else Executer::PreDecrement(data); break; 

			case Misc::ExecFuncs::If: Executer::If(data); break;
			case Misc::ExecFuncs::ElseIf: Executer::If(data); break; // same as if
			case Misc::ExecFuncs::Else: Executer::Else(data); break;

			case Misc::ExecFuncs::For: Executer::For(data); break;
			case Misc::ExecFuncs::ForEach: Executer::ForEach(data); break;
			case Misc::ExecFuncs::While: Executer::While(data); break;
				
			case Misc::ExecFuncs::New: return Executer::New(data); break;
			case Misc::ExecFuncs::Class: Executer::CreateClass(data); break;
			case Misc::ExecFuncs::Function: Executer::CreateFunction(data); break;
			case Misc::ExecFuncs::AnonFunction: return Executer::CreateAnonFunction(data); break;

			case Misc::ExecFuncs::ExecuteFunction:
				if (data->FuncCallback) return Executer::ExecuteFunction(data);
				else Executer::ExecuteFunction(data);
				break; 

			case Misc::ExecFuncs::End:
				return data->FuncCallback ? data->BromScript->GC.RegisterVariable() : null; // lets give back null value.

			case Misc::ExecFuncs::MergeStart: return Executer::Merge(data);
			case Misc::ExecFuncs::Return: return Executer::Return(data);
			case Misc::ExecFuncs::Break: data->Breaking = true; return data->FuncCallback ? data->BromScript->GC.RegisterVariable() : null;
			case Misc::ExecFuncs::Continue: data->Continueing = true; return data->FuncCallback ? data->BromScript->GC.RegisterVariable() : null;
				
			case Misc::ExecFuncs::GetCount: return Executer::GetCount(data);
			case Misc::ExecFuncs::Bool: return Executer::GetBool(data);
			case Misc::ExecFuncs::String: return Executer::GetString(data);
			case Misc::ExecFuncs::Number: return Executer::GetNumber(data);
			case Misc::ExecFuncs::Table: return Executer::GetTable(data);
			case Misc::ExecFuncs::Enum: return Executer::GetEnum(data);
				
			case Misc::ExecFuncs::Skip: break;
			case Misc::ExecFuncs::StringTable: Executer::StringTable(data); break;
			case Misc::ExecFuncs::CurrentLine:
				this->CurrentSourceFileLine = data->Reader->ReadInt();
				this->BromScript->Debug->Update();
				break;

			default:
				data->BromScript->Error(CString::Format("Invalid bytecode detected! (%d)", b));

				data->HasReturnValue = true;
				data->Returning = data->BromScript->GC.RegisterVariable();
				return data->Returning;
			}
		}
	}
}